#include <avr/io.h>
#include "mmc_lib.h"
#include "fat16.h"
#include <string.h>

uint8_t fat_buf[512];

static uint16_t  RootDirRegionStartSec;
static uint32_t  DataRegionStartSec;
static uint16_t  RootDirRegionSize;
static uint8_t   SectorsPerCluster;
static uint16_t  FATRegionStartSec;

unsigned char fat16_init(void)
{
	mbr_t *mbr = (mbr_t*) fat_buf;
	vbr_t *vbr = (vbr_t*) fat_buf;
	
	if (mmc_init() != MMC_OK)
		return 1;
	
	//Load MBR
	//mmc_set_blocklen(512);
	mmc_start_read_block(0, 9);
//	mmc_read_buffer((unsigned char *) mbr, 512);
	mmc_read_buffer();
	mmc_stop_read_block();
	
	if ((mbr->sector.partition[0].typeId != PARTTYPE_FAT16L) || (mbr->sector.magic != 0xAA55))
		return 2;
		
	FATRegionStartSec = mbr->sector.partition[0].sectorOffset;
	
	//Load VBR
	mmc_start_read_block(mbr->sector.partition[0].sectorOffset, 9);
	mmc_read_buffer();
	mmc_stop_read_block();
	
	SectorsPerCluster  			= vbr->bsSecPerClus;		// 4
	
	// Calculation Algorithms
	FATRegionStartSec			+= vbr->bsRsvdSecCnt;						// 6
	RootDirRegionStartSec	 	= FATRegionStartSec + (vbr->bsNumFATs * vbr->bsNrSeProFAT16);		// 496	
//	RootDirRegionSize		 	= (vbr->bsRootEntCnt * 32) / 512;						// 32
	RootDirRegionSize		 	= (vbr->bsRootEntCnt / 16); 						// 32
	DataRegionStartSec 			= RootDirRegionStartSec + RootDirRegionSize;	// 528
	
	return 0;
}

uint16_t fat16_readRootDirEntry(uint16_t entry_num, uint32_t *size, uint8_t *attrib)
{
	direntry_t *dir; //Zeiger auf einen Verzeichniseintrag
	
//	if ((entry_num * sizeof(direntry_t) / 512) >= RootDirRegionSize)
	if ((entry_num / 16) >= RootDirRegionSize)
		return 0xFFFF; //End of root dir region reached!
	
//	uint32_t dirsector = RootDirRegionStartSec + entry_num * sizeof(direntry_t) / 512;
	uint32_t dirsector = RootDirRegionStartSec + entry_num / 16;
	entry_num %= 512 / sizeof(direntry_t);
	
	mmc_start_read_block(dirsector, 9);
	mmc_read_buffer();
	mmc_stop_read_block();

	dir = (direntry_t *) fat_buf + entry_num;

	if ((dir->name[0] == 0) || (dir->name[0] == 0xE5) || (dir->fstclust == 0))
		return 0xFFFF;

	*size = dir->filesize;
	*attrib = dir->attr;

	return dir->fstclust;
}


void fat16_readfilesector(uint16_t startcluster, uint32_t filesector)
{
	uint16_t clusteroffset;
	uint16_t currentfatsector;
	uint8_t temp, secoffset;
	uint32_t templong;
	
	fatsector_t *fatsector = (fatsector_t*) fat_buf;
	
         // SectorsPerCluster is always power of 2 !
         //clusteroffset = filesector; // Upps, das geht natrlich nicht !
         templong = filesector;
         if(SectorsPerCluster>1)
          {
//          clusteroffset = filesector / SectorsPerCluster;
           temp = SectorsPerCluster >> 1;
           while(temp)
            {
             templong >>= 1;
             temp >>= 1;
            }
          }

         clusteroffset = (uint16_t)templong;

//	filesector %= SectorsPerCluster;
//	filesector &= (SectorsPerCluster-1); // SectorsPerCluster is always power of 2 !
        // SectorsPerCluster is 128 max. so result is uint8 !
	secoffset = (uint8_t)filesector & (SectorsPerCluster-1); // SectorsPerCluster is always power of 2 !
	
	currentfatsector = 0xFFFF;
	while (clusteroffset)
	{
		if (currentfatsector != (startcluster / 256))
		{
			mmc_start_read_block(FATRegionStartSec + startcluster / 256, 9);
			mmc_read_buffer();
			mmc_stop_read_block();
			currentfatsector = startcluster / 256;
		}
		
		startcluster = fatsector->fat_entry[startcluster % 256];
		clusteroffset--;
	}
	
//	if ((startcluster == 0) | (startcluster == 1))
//	if ((startcluster == 0) || (startcluster == 1))
	if (startcluster < 2)
		return; //Free/reserved Sector
	
//	if ((startcluster & 0xFFFFFFF0) == 0xFFFFFFF0)
	if ((startcluster & 0xFFF0) == 0xFFF0) //FAT16 hat nur 16Bit Clusternummern
		return; //Reserved / bad / last cluster

//        (unsigned long) (startcluster - 2) * SectorsPerCluster
// SectorsPerCluster is always power of 2 !
        templong = startcluster - 2;
        if(SectorsPerCluster>1)
         {
          temp = SectorsPerCluster>>1;
          while(temp)
           {
            templong <<= 1;
            temp >>= 1;
           }
         }
		
//	mmc_start_read_block((unsigned long) (startcluster - 2) * SectorsPerCluster + DataRegionStartSec + filesector, 9);
	mmc_start_read_block(templong + DataRegionStartSec + secoffset, 9);
	mmc_read_buffer();
	mmc_stop_read_block();
}
